#!/usr/bin/python3
"""
## © University of Southampton IT Innovation Centre, 2018
##
## Copyright in this software belongs to University of Southampton
## IT Innovation Centre of Gamma House, Enterprise Road,
## Chilworth Science Park, Southampton, SO16 7NS, UK.
##
## This software may not be used, sold, licensed, transferred, copied
## or reproduced in whole or in part in any manner or form or in or
## on any media by any person other than in accordance with the terms
## of the Licence Agreement supplied with the software, or otherwise
## without the prior written consent of the copyright owners.
##
## This software is distributed WITHOUT ANY WARRANTY, without even the
## implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
## PURPOSE, except where stated in the Licence Agreement supplied with
## the software.
##
##      Created By :            Nikolay Stanchev
##      Created Date :          25-04-2018
##      Created for Project :   FLAME
"""

from influxdb import InfluxDBClient
from time import time, sleep
from urllib.parse import urlparse
from threading import Thread, Event
import clmctest.monitoring.LineProtocolGenerator as lp


class Aggregator(Thread):
    """
    A class used to perform the aggregation feature of the CLMC - aggregation network and media service measurements. Currently, implemented as a thread,
    so that the implementation can be tested using pytest.
    """
    
    PATH_SEPARATOR = "---"  # currently, the path id is assumed to be in the format "SourceEndpointID---TargetEndpointID"
    REPORT_PERIOD = 5  # currently, report period is 5s, that is every 5 seconds the mean delay values for the last 5 seconds are aggregated
    DATABASE = 'E2EMetrics'  # default database the aggregator uses
    DATABASE_URL = 'http://203.0.113.100:8086'  # default database URL the aggregator uses
    
    def __init__(self, database=DATABASE, database_url=DATABASE_URL):
        """
        Constructs an Aggregator instance.

        :param database: database name to use
        :param database_url: database url to use
        """

        super(Aggregator, self).__init__()  # call the constructor of the tread

        # initialise a database client using the database url and the database name
        url_object = urlparse(database_url)
        self.db_client = InfluxDBClient(host=url_object.hostname, port=url_object.port, database=database, timeout=10)

        self.db_url = database_url
        self.db_name = database

        # a stop flag event object used to handle the killing of the thread
        self._stop_flag = Event()

    def stop(self):
        """
        A method used to stop the thread.
        """

        self._stop_flag.set()

    def run(self):
        """
        Performs the functionality of the aggregator - query data from both measurements merge that data and post it back in influx every 5 seconds.
        """

        current_time = int(time())
        while True:
            if self._stop_flag.is_set():
                break

            boundary_time = current_time - Aggregator.REPORT_PERIOD

            boundary_time_nano = boundary_time * 1000000000
            current_time_nano = current_time * 1000000000

            network_delays = {}
            result = self.db_client.query(
                'SELECT mean(delay) as "Dnet" FROM "E2EMetrics"."autogen".network_delays WHERE time >= {0} and time < {1} GROUP BY path'.format(
                    boundary_time_nano, current_time_nano))
            for item in result.items():
                metadata, result_points = item
                # measurement = metadata[0]
                tags = metadata[1]

                network_delays[tags['path']] = next(result_points)['Dnet']

            service_delays = {}
            result = self.db_client.query(
                'SELECT mean(response_time) as "Dresponse" FROM "E2EMetrics"."autogen".service_delays WHERE time >= {0} and time < {1} GROUP BY FQDN, sf_instance, endpoint'.format(
                    boundary_time_nano, current_time_nano))
            for item in result.items():
                metadata, result_points = item
                # measurement = metadata[0]
                tags = metadata[1]

                if tags['endpoint'] not in service_delays:
                    service_delays[tags['endpoint']] = [(next(result_points)['Dresponse'], tags['FQDN'], tags['sf_instance'])]
                else:
                    service_delays[tags['endpoint']].append((next(result_points)['Dresponse'], tags['FQDN'], tags['sf_instnace']))

            for path in network_delays:
                target_endpoint = self.get_target_endpoint(path)
                if target_endpoint not in service_delays:
                    continue

                e2e_arguments = {"path_id_f": None, "path_id_r": None, "fqdn": None, "sf_instance": None, "delay_path_f": None, "delay_path_r": None,
                                 "delay_service": None, "time": boundary_time}

                e2e_arguments['path_id_f'] = path
                e2e_arguments['delay_path_f'] = network_delays[path]

                reversed_path = self.reverse_path_id(path)
                assert reversed_path in network_delays
                e2e_arguments['path_id_r'] = reversed_path
                e2e_arguments['delay_path_r'] = network_delays[reversed_path]

                for service_delay in service_delays[target_endpoint]:
                    response_time, fqdn, sf_instance = service_delay
                    e2e_arguments['delay_service'] = response_time
                    e2e_arguments['fqdn'] = fqdn
                    e2e_arguments['sf_isntnace'] = sf_instance

                    if None not in e2e_arguments.items():
                        self.db_client.write_points(
                            lp.generate_e2e_delay_report(e2e_arguments['path_id_f'], e2e_arguments['path_id_r'], e2e_arguments['fqdn'], e2e_arguments['sf_isntnace'],
                                                         e2e_arguments['delay_path_f'], e2e_arguments['delay_path_r'], e2e_arguments['delay_service'], e2e_arguments['time']))

            old_timestamp = current_time
            while current_time != old_timestamp + 5:
                sleep(1)
                current_time = int(time())
    
    @staticmethod
    def reverse_path_id(path):
        """
        Reverses a path identifier.
        :param path: the path ID assumed to be in format "SourceEndpointID---TargetEndpointID"
        :return: the reversed path ID, e.g. "TargetEndpointID---SourceEndpointID"
        """

        source_endpoint, target_endpoint = path.split(Aggregator.PATH_SEPARATOR)
        return "{0}{1}{2}".format(target_endpoint, Aggregator.PATH_SEPARATOR, source_endpoint)
    
    @staticmethod
    def get_target_endpoint(path):
        """
        Get a target endpoint by parsing a path identifier.

        :param path: the path ID assumed to be in format "SourceEndpointID---TargetEndpointID"
        :return: the target endpoint retrieved from the path identifier.
        """

        return path.split(Aggregator.PATH_SEPARATOR)[1]


if __name__ == '__main__':
    Aggregator().start()
